                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--

*/

#include <afxwin.h>
#include <afxtempl.h>

#include <d3d8.h>
#include <d3dx8.h>

D3DCAPS8 caps;
LPDIRECT3D8       d3d       = NULL;
LPDIRECT3DDEVICE8 d3ddevice = NULL;

#include "Common/CPU.h"

#include "DirectX/Texture.h"
#include "3D/Resolution.h"

#include "3D/PerlinNoise.h"

#define VK_ENTER 13



#define WATER_SCALE      1.0f
#define WAVE_SCALE       0.0524f
#define WATER_SIZE       128
#define WATER_HALF_SIZE (WATER_SIZE >> 1)

struct WaterVertex {
	Vertex pos;
	float u, v;
	Vertex reflectionVector;
};

/* ################################################################################################################### */

class Direct3DWindow : public CFrameWnd {
public:
    Direct3DWindow(HWND hwnd);

	void DrawSkyBox();
	void DrawWater(float time);
	void DrawFrame();
protected:
	CRect windowedRect;
	int Width, Height;
	int Refresh;
	bool vSync;
	bool windowMoved;

    __int64 cpuHz;

	TextureHandler *texHandler;
	TextureID water, cubeMap, proceduralTexture;

	Vertex position;

	ShaderID fragmentShader;

	BOOL PreCreateWindow(CREATESTRUCT &cs);
	int  OnCreate(LPCREATESTRUCT lpCreateStruct);
	void OnSize(UINT nType, int cx, int cy);
	void OnClose();

	void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	void OnMouseMove  (UINT nFlags, CPoint point);

	void OnKillFocus(CWnd *pNewWnd);

	DECLARE_MESSAGE_MAP ()
};

BEGIN_MESSAGE_MAP(Direct3DWindow, CFrameWnd)
	
	ON_WM_CREATE ()
	ON_WM_CLOSE ()

	ON_WM_KEYDOWN ()
	ON_WM_MOUSEMOVE ()
	ON_WM_KILLFOCUS ()

END_MESSAGE_MAP()


/* ################################################################################################################### */

void Direct3DWindow::DrawSkyBox(){
	texHandler->setTextures(cubeMap);
	texHandler->apply();

	static float vertices[] = {
		-1.0f,  1.0f,  1.0f, -1.0f,  1.0f,  1.0f,
		 1.0f,  1.0f,  1.0f,  1.0f,  1.0f,  1.0f,
		-1.0f, -1.0f,  1.0f, -1.0f, -1.0f,  1.0f,
		 1.0f, -1.0f,  1.0f,  1.0f, -1.0f,  1.0f,

		-1.0f,  1.0f, -1.0f, -1.0f,  1.0f, -1.0f,
		 1.0f,  1.0f, -1.0f,  1.0f,  1.0f, -1.0f,
		-1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f,
		 1.0f, -1.0f, -1.0f,  1.0f, -1.0f, -1.0f,
	};

	static int indices[] = {0,1,2,  1,2,3,  1,5,3,  5,3,7,  5,4,7,  4,6,7,  4,0,6,  0,6,2,  4,5,0,  5,0,1,  2,3,6,  3,6,7};

	d3ddevice->SetVertexShader(D3DFVF_XYZ | D3DFVF_TEXCOORDSIZE3(0) | (1 << D3DFVF_TEXCOUNT_SHIFT));
	d3ddevice->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST, 0, sizeof(indices) / sizeof(int), 12, indices, D3DFMT_INDEX32, vertices, 6 * sizeof(float));
}

void Direct3DWindow::DrawWater(float time){
	WaterVertex *dest;
	float x,z;
	int i,j;

	static WaterVertex waterVertices[WATER_SIZE][WATER_SIZE];
	int indices[2 * WATER_SIZE];
	
	static bool first = true;
	if (first){
		dest = waterVertices[0];
		for (z = -WATER_HALF_SIZE; z < WATER_HALF_SIZE; z++){
			for (x = -WATER_HALF_SIZE; x < WATER_HALF_SIZE; x++){
				dest->pos.x = x * WATER_SCALE * 2;
				dest->pos.z = z * WATER_SCALE;
				dest->u = 9 * (x + WATER_HALF_SIZE) / WATER_SIZE;
				dest->v = 7 * (z + WATER_HALF_SIZE) / WATER_SIZE;
				dest++;
			}
		}

		first = false;
	}

	// Debug
	dest = waterVertices[0];
	for (z = -WATER_HALF_SIZE; z < WATER_HALF_SIZE; z++){
		for (x = -WATER_HALF_SIZE; x < WATER_HALF_SIZE; x++){
			dest->v += 0.003f;
			dest++;
		}
	}

	time *= 0.3f;

	dest = waterVertices[0];
	for (z = -WATER_HALF_SIZE; z < WATER_HALF_SIZE; z++){
		for (x = -WATER_HALF_SIZE; x < WATER_HALF_SIZE; x++){
			dest->pos.y = 0.5f * (1.0f + noise3(0.25f * x * WAVE_SCALE, 17 * time * WAVE_SCALE, (z + 24 * time) * WAVE_SCALE));
			dest++;
		}
	}


	#define succ(x) (x + 1) - (x == WATER_SIZE - 1)
	#define pred(x) (x - 1) + (x == 0)

	// Apply Sobel filter to find mesh normals
	Vertex normal, eyeVector;
	dest = waterVertices[0];
	for (j = 0; j < WATER_SIZE; j++){
		for (i = 0; i < WATER_SIZE; i++){
			normal.x =
				     waterVertices[pred(j)][pred(i)].pos.y  -
				     waterVertices[pred(j)][succ(i)].pos.y  +
				2 * (waterVertices[     j ][pred(i)].pos.y  -
				     waterVertices[     j ][succ(i)].pos.y) +
					 waterVertices[succ(j)][pred(i)].pos.y  -
					 waterVertices[succ(j)][succ(i)].pos.y;

			normal.y = 1.0f;

			normal.z =      
				     waterVertices[pred(j)][pred(i)].pos.y  -
				     waterVertices[succ(j)][pred(i)].pos.y  +
				2 * (waterVertices[pred(j)][     i ].pos.y  -
				     waterVertices[succ(j)][     i ].pos.y) +
					 waterVertices[pred(j)][succ(i)].pos.y  -
					 waterVertices[succ(j)][succ(i)].pos.y;

			normal.fastNormalize();
	
			eyeVector = position - dest->pos;
			eyeVector.fastNormalize();

			dest->reflectionVector = 2 * (eyeVector * normal) * normal - eyeVector;

			dest++;
		}
	}

	for (i = 0; i < WATER_SIZE; i++){
		indices[2 * i    ] = i;
		indices[2 * i + 1] = i + WATER_SIZE;
	}

	texHandler->setTextures(proceduralTexture, cubeMap, water);
	texHandler->setFragmentShader(fragmentShader);
	texHandler->apply();

	d3ddevice->SetVertexShader(
		D3DFVF_XYZ | 
		D3DFVF_TEXCOORDSIZE2(0) |
		D3DFVF_TEXCOORDSIZE3(1) |
		(2 << D3DFVF_TEXCOUNT_SHIFT)
	);

	for (i = 0; i < WATER_SIZE - 1; i++){
		d3ddevice->DrawIndexedPrimitiveUP(D3DPT_TRIANGLESTRIP, i * WATER_SIZE, WATER_SIZE * 2, WATER_SIZE * 2 - 2, indices, D3DFMT_INDEX32, waterVertices, sizeof(WaterVertex));

		for (j = 0; j < WATER_SIZE * 2; j++) indices[j] += WATER_SIZE;
	}

}

void Direct3DWindow::DrawFrame(){
    static bool first = true;

    static __int64 prev, curr;
    static float currTime;
    float time;

    prev = curr;
    curr = getCycleNumber();
    if (first){
        prev = curr;
        first = false;
    }
    time = (float) (double(curr - prev) / cpuHz);
    currTime += time;



    d3ddevice->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    d3ddevice->BeginScene();


	D3DXMATRIX modelView, projection, tmp;
	
	D3DXMatrixPerspectiveFovLH(&projection, 80 * PI / 180, (float) Width / (float) Height, 0.01f, 500);
	d3ddevice->SetTransform(D3DTS_PROJECTION, &projection);

	D3DXMatrixIdentity(&tmp);
	d3ddevice->SetTransform(D3DTS_WORLD, &tmp);
	d3ddevice->SetTransform(D3DTS_VIEW, &tmp);

	d3ddevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);

	DrawSkyBox();

	d3ddevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);

	D3DXMatrixTranslation(&modelView, -position.x, -position.y, -position.z);
	d3ddevice->SetTransform(D3DTS_VIEW, &modelView);


    
	texHandler->update(proceduralTexture, (256 * curr) / (cpuHz / 1024));
	
	DrawWater(currTime);
	
    d3ddevice->EndScene();

    d3ddevice->Present(NULL, NULL, NULL, NULL);
}


Direct3DWindow::Direct3DWindow(HWND hwnd){
	Width   = GetSystemMetrics(SM_CXSCREEN);
	Height  = GetSystemMetrics(SM_CYSCREEN);
	Refresh = 85;
	vSync = false;

	position = Vertex(0,8,-WATER_HALF_SIZE * WATER_SCALE);

	CRect r(0, 0, Width, Height);
	windowedRect = r;

	Create(NULL, "Direct3D", WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, r, hwnd == NULL? NULL : FromHandle(hwnd));
}

BOOL Direct3DWindow::PreCreateWindow(CREATESTRUCT &cs){
	CFrameWnd::PreCreateWindow(cs);

	cs.dwExStyle &= ~WS_EX_CLIENTEDGE;

	return 1;
}

int Direct3DWindow::OnCreate(LPCREATESTRUCT lpCreateStruct){
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1) return -1;

	ShowWindow(SW_SHOW);
	SetForegroundWindow();
	SetFocus();

    initCPU();
    cpuHz = getHz();

	texHandler = NULL;
	if ((d3d = Direct3DCreate8(D3D_SDK_VERSION)) == NULL){
		MessageBox("Couldn't initialize Direct3D\nMake sure you have installed DirectX 8.1 or later.", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}
	

	d3d->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &caps);

	if (!(caps.TextureCaps & D3DPTEXTURECAPS_CUBEMAP)){
		MessageBox("No support for cubemaps!", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}

	// Pixel shaders 1.4
	if ((caps.PixelShaderVersion & 0xFFFF) < ((1 << 8) | 4)){
		MessageBox("This demo requires pixel shader 1.4 or higher", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}


    D3DPRESENT_PARAMETERS d3dpp; 
    ZeroMemory(&d3dpp, sizeof(d3dpp));

	D3DFORMAT formats[] = {D3DFMT_A8R8G8B8, D3DFMT_X8R8G8B8};
	if (!d3dFindBestResolution(Width, Height, formats, sizeof(formats) / sizeof(D3DFORMAT), Refresh)){
		MessageBox("No suitable displaymode found.", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
		return 0;
	}
	
	d3dpp.Windowed = FALSE;
	d3dpp.BackBufferWidth  = Width;
	d3dpp.BackBufferHeight = Height;
	d3dpp.BackBufferFormat = formats[0];
	
	d3dpp.FullScreen_RefreshRateInHz = Refresh;
	if (vSync){
		d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_ONE;
		d3dpp.BackBufferCount = 2;
	} else {
		d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		d3dpp.BackBufferCount = 1;
	}
	d3dpp.EnableAutoDepthStencil = TRUE;
	d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;

	checkTextureFormats(d3dpp.BackBufferFormat);
	
    if (FAILED(d3d->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, (HWND) *this, /*D3DCREATE_SOFTWARE_VERTEXPROCESSING*/ D3DCREATE_MIXED_VERTEXPROCESSING /*D3DCREATE_HARDWARE_VERTEXPROCESSING*/, &d3dpp, &d3ddevice))){
		MessageBox("Couldn't create Direct3D device interface.", "Error", MB_OK | MB_ICONERROR);
		PostMessage(WM_CLOSE);
        return 0;
    }
	


	// Disable some odd defaults
	d3ddevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	d3ddevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);


	texHandler = new TextureHandler();

	fragmentShader = texHandler->createFragmentShader(
	"\
          ps.1.4                                \
											    \
		  def c0, 0.02f, 0.02f, 0.02f, 0.0f     \
		  def c1, 0.8f, 0.8f, 0.0f, 0.0f        \
		  def c2, 0.2f, 0.2f, 0.2f, 0.0f        \
											    \
		  texld   r0, t0                        \
		  texcrd  r1.rgb, t1                    \
		  texcrd  r2.rgb, t2                    \
		                                        \
		  mul     r0, r0_bias, c0               \
		  add     r1.rgb, r0, r1                \
		  add     r2.rgb, r0, r2                \
		                                        \
          phase                                 \
		                                        \
		  texld   r1, r1                        \
		  texld   r2, r2                        \
		                                        \
		  mul     r2, r2, c2                    \
		  dp3_sat r0, r1, c1                    \
		  lrp     r0, r0, r1, r2                \
		");

	cubeMap  = texHandler->loadCubeMapFromFiles(
		"Sea_Textures\\sun_posx.jpg",
		"Sea_Textures\\sun_negx.jpg",
		"Sea_Textures\\sun_posy.jpg",
		"Sea_Textures\\sun_negy.jpg",
		"Sea_Textures\\sun_posz.jpg",
		"Sea_Textures\\sun_negz.jpg");

	water    = texHandler->loadFromFile("Sea_Textures\\Water.jpg");

    proceduralTexture = texHandler->createProceduralTexture(MakePerlinNoiseTexture);

	texHandler->uploadAll();

	
	ShowCursor(FALSE);

	initPerlin(37);

	return 0;
}

void Direct3DWindow::OnSize(UINT nType, int cx, int cy){
	D3DVIEWPORT8 vp;
	vp.X      = 0;
	vp.Y      = 0;
	vp.Width  = cx;
	vp.Height = cy;
	vp.MinZ   = 0.0f;
	vp.MaxZ   = 1.0f;
	d3ddevice->SetViewport(&vp);
}

void Direct3DWindow::OnClose(){
	if (texHandler != NULL){

		texHandler->unloadAll();
		delete texHandler;
	}

    if (d3ddevice != NULL) d3ddevice->Release();
    if (d3d != NULL) d3d->Release();

	ShowCursor(TRUE);
	DestroyWindow();
}


void Direct3DWindow::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags){
	PostMessage(WM_CLOSE, 0,0);
}

void Direct3DWindow::OnMouseMove(UINT nFlags, CPoint point){
	static __int64 first = 0, last;
	__int64 curr;

	if (first == 0){
		first = last = getCycleNumber();
	} else {
		curr = getCycleNumber();
		if (curr - first > cpuHz && curr - last < cpuHz / 8){
			PostMessage(WM_CLOSE, 0,0);
		}
		last = curr;
	}
}

void Direct3DWindow::OnKillFocus(CWnd *pNewWnd){
	PostMessage(WM_CLOSE, 0,0);
}

/* ################################################################################################################### */


class Direct3DApp : public CWinApp {
public:
    virtual BOOL InitInstance();
	BOOL OnIdle(LONG lCount);
};


BOOL Direct3DApp::InitInstance(){
	// Fix current path
	char dir[256];
	GetModuleFileName(NULL, dir, sizeof(dir));
	*strrchr(dir, '\\') = '\0';
	SetCurrentDirectory(dir);

	//MessageBox(GetActiveWindow(), m_lpCmdLine, "Info", MB_OK | MB_ICONINFORMATION);

	if (strncmp(m_lpCmdLine, "/c", 2) == 0){
		MessageBox(GetActiveWindow(), "A screen saver by Humus.\n\nhttp://esprit.campus.luth.se/~humus/", "Info", MB_OK | MB_ICONINFORMATION);
	} else if (strncmp(m_lpCmdLine, "/p", 2) == 0){
		// Never got this to work :/
		return FALSE;
	} else if (strncmp(m_lpCmdLine, "/s", 2) == 0){
		m_pMainWnd = new Direct3DWindow(NULL);
		return TRUE;
	}

	return FALSE;
}

BOOL Direct3DApp::OnIdle(LONG lCount){
	
	((Direct3DWindow *) m_pMainWnd)->DrawFrame();

	return TRUE;
}

Direct3DApp Direct3DApp;


/* ########################################## End of file ############################################################ */
